"""
The command to exit snadra.
"""
from typing import TYPE_CHECKING

from snadra._core.base import Complete, Parameter
import snadra._utils as snutils
from snadra._utils import CommandMeta

if TYPE_CHECKING:
    import argparse


class Command(CommandMeta):
    """
    Help message for "exit".
    """

    keywords = {"exit", "quit"}
    description = "Exit the console"
    long_help = "LONG HELP FOR EXIT COMMAND"

    arguments = {
        "-y,--yes": Parameter(
            complete=Complete.NONE, action="store_true", help="Confirm to exit"
        )
    }

    def run(self, args: "argparse.Namespace") -> None:
        """
        Exit `snadra`.

        Parameters
        ----------
        args : :class:`argparse.Namespace`
            The arguments for the command.
        """
        if not args.yes:
            snutils.console.log("[red]Error[/red]: Exit not confirmed (use '--yes')")
            return

        raise EOFError("Got an exit signal")



import argparse
import enum
import functools
from typing import Dict, Set

import pygments.token as ptoken


class Complete(enum.Enum):
    """
    Command arguments, completion options.

    Attributes
    ----------
    CHOICES : enum.auto
        Complete argument from the list of choices specified in `parameter`.
    NONE : enum.auto
        Do not provide argument completions.
    """

    CHOICES = enum.auto()
    NONE = enum.auto()


class Parameter:
    """
    Representation of a parameter for the command parsing.

    Attributes
    ----------
    complete : snadra._core.base.Complete
    token : pygments:token.Name.Label
    group : str, optional
    """

    def __init__(
        self,
        complete: Complete,
        token=ptoken.Name.Label,
        *args,
        **kwargs,
    ) -> None:
        self.complete = complete
        self.token = token
        self.args = args
        self.kwargs = kwargs


class CommandDefinition:
    """
    The generic structure for commands.

    Attributes
    ----------
    keywords : set of str
        Set of the keywords for the new command.
    description : str
        Help text for the new command.
    long_help : str
        Long help for the command.
    arguments : Dict[str, snadra._core.base.Parameter]
        Dictionary of parameter definitions created with the :class:`Parameter` class.
        If this is None, your command will receive the
        raw argument string and no processing will be done except
        removing the leading command name.
    """

    keywords: Set[str] = {"unimplemented"}
    description: str = ""
    long_help: str = ""
    arguments: Dict[str, Parameter] = {}
    defaults: Dict = {}

    def __init__(self) -> None:
        # Create the parser object
        if self.arguments is not None:
            for keyword in self.keywords:
                self.parser = argparse.ArgumentParser(
                    prog=keyword,
                    description=self.description,
                    formatter_class=argparse.RawDescriptionHelpFormatter,
                )
                self.build_parser(self.parser, self.arguments)
        else:
            self.parser = None  # type: ignore

    def run(self, args: argparse.Namespace):
        """
        This is what gets run for each command.

        Parameters
        ----------
        args : argparse.Namespace
            Namespace containing the parsed arguments.

        Raises
        ------
        NotImplementedError
            If there was no :meth:`snadra._core.base.CommandDefinition.run`
            method for the new command's class.
        """
        raise NotImplementedError

    def build_parser(
        self,
        parser: argparse.ArgumentParser,
        args: Dict[str, Parameter],
    ) -> None:
        """
        Parse the ARGS and DEFAULTS dictionaries to build an argparse ArgumentParser
        for this command. You should not need to overload this.

        Parameters
        ----------
        parser : argparse.ArgumentParser
            Parser object to add arguments to.
        args : Dict[str, snadra._core.base.Parameter]
            `ARGS` dictionary.
        """
        for arg, param in args.items():
            names = arg.split(",")
            group = parser

            # Patch choice to work with a callable
            if "choices" in param.kwargs and callable(param.kwargs["choices"]):
                method = param.kwargs["choices"]

                class wrapper:
                    def __init__(wself, method) -> None:
                        wself.method = method

                    def __iter__(wself):
                        yield from wself.method(self)

                param.kwargs["choices"] = wrapper(method)

            # Patch "type" so we can see "self"
            if (
                "type" in param.kwargs
                and isinstance(param.kwargs["type"], tuple)
                and param.kwargs["type"][0] == "method"
            ):
                param.kwargs["type"] = functools.partial(param.kwargs["type"][1], self)

            group.add_argument(*names, *param.args, **param.kwargs)

        parser.set_defaults(**self.defaults)



import abc
import argparse
import functools
from typing import Dict, Set

from rich.console import Console

from snadra._core.base import Parameter

console = Console()


class CommandMeta(metaclass=abc.ABCMeta):
    """
    Abstract base class for command line commands.

    See Also
    --------
    snadra._core.commands.exit
    snadra._core.commands.help
    """

    def __init__(self) -> None:
        # Create the parser object
        if self.arguments is not None:
            for keyword in self.keywords:
                self.parser = argparse.ArgumentParser(
                    prog=keyword,
                    description=self.description,
                    formatter_class=argparse.RawDescriptionHelpFormatter,
                )
                self.build_parser(self.parser, self.arguments)
        else:
            self.parser = None  # type: ignore

    def build_parser(
        self,
        parser: argparse.ArgumentParser,
        args: Dict[str, Parameter],
    ) -> None:
        """
        Parse the ARGS and DEFAULTS dictionaries to build an argparse ArgumentParser
        for this command. You should not need to overload this.

        Parameters
        ----------
        parser : argparse.ArgumentParser
            Parser object to add arguments to.
        args : Dict[str, snadra._core.base.Parameter]
            `ARGS` dictionary.
        """
        for arg, param in args.items():
            names = arg.split(",")
            group = parser

            # Patch choice to work with a callable
            if "choices" in param.kwargs and callable(param.kwargs["choices"]):
                method = param.kwargs["choices"]

                class wrapper:
                    def __init__(wself, method) -> None:
                        wself.method = method

                    def __iter__(wself):
                        yield from wself.method(self)

                param.kwargs["choices"] = wrapper(method)

            # Patch "type" so we can see "self"
            if (
                "type" in param.kwargs
                and isinstance(param.kwargs["type"], tuple)
                and param.kwargs["type"][0] == "method"
            ):
                param.kwargs["type"] = functools.partial(param.kwargs["type"][1], self)

            group.add_argument(*names, *param.args, **param.kwargs)

        parser.set_defaults(**self.defaults)

    @property
    def defaults(self):
        """
        foo bar baz.
        """
        return {}

    @property
    @abc.abstractmethod
    def keywords(self) -> Set[str]:
        """
        Keywords for the new command.

        Can be treated as command "aliases".

        Returns
        -------
        set of str
            Keywords for the command.
        """
        ...

    @property
    @abc.abstractmethod
    def description(self) -> str:
        """
        Command description.

        Returns
        -------
        str
            The description for the command.
        """
        ...

    @property
    @abc.abstractmethod
    def long_help(self) -> str:
        """
        Long help text for the new command.

        Returns
        -------
        str
            Long help text for the command.
        """
        ...

    @property
    @abc.abstractmethod
    def arguments(self) -> Dict[str, "Parameter"]:
        """
        Arguments for the new command.

        Dictionary of parameter definitions created with the `Parameter` class.
        If this is None, your command will receive the
        raw argument string and no processing will be done except
        removing the leading command name.

        See Also
        --------
        snadra._core.base.Parameter
        """
        ...

    @abc.abstractmethod
    def run(self, args: argparse.Namespace) -> None:
        """
        Implementation for the new command.

        This is responsible for the actual action of the command.
        """
        ...



from typing import TYPE_CHECKING, Set

from rich import box as rich_box
from rich.table import Table as RichTable

from snadra._core.base import Complete, Parameter
from snadra._core.commands import Commands
import snadra._utils as snutils
from snadra._utils import CommandMeta

if TYPE_CHECKING:
    import argparse


def _all_available_keywords(commands: Commands, help_keywords: Set[str]) -> Set[str]:
    """
    All the available keywords, of the core commands.

    Parameters
    ----------
    commands : snadra._core.commands.Commands
        Commands objecs.
    help_keywords : set of str
        Keywords of the `help` command.

    Returns
    -------
    set of str
        All the available, core command's keywords.

    See Also
    --------
    snadra._core.commands.Commands
    """
    commands_keywords = commands.keywords
    all_available_keywords = commands_keywords.union(help_keywords)
    return all_available_keywords


class Command(CommandMeta):
    """
    The command `help`, for displaying help information about other commands.
    """

    # TODO: Add tests

    keywords = {"help"}
    description = "List all known commands and print their help message"
    long_help = "THE LONG HELP MESSAGE OF 'help'"

    # TODO: Get rid of this 'choices' and make better handling and
    # display better information about the commands and how to use the help command.
    _core_commands = Commands(
        skip={
            "__init__",
            "help",  # Skipping "help" so we won't make a circular import
        }
    )
    _available_keywords = sorted(
        _all_available_keywords(commands=_core_commands, help_keywords=keywords)
    )
    arguments = {
        "topic": Parameter(
            Complete.CHOICES,
            choices=_available_keywords,
            nargs="?",
        )
    }

    def run(self, args: "argparse.Namespace") -> None:
        """
        Show the help.

        Parameters
        ----------
        args : :class:`argparse.Namespace`
            The arguments for the command.
        """
        if args.topic:
            if args.topic in self.keywords:
                snutils.console.print(self.long_help)
            else:
                # Here we are counting on "argparse" choices for validation.
                target_command = self._core_commands.get_command(args.topic)
                snutils.console.print(target_command.long_help)  # type: ignore

        elif args:
            help_table = RichTable(title="Help menu", box=rich_box.SIMPLE)
            help_table.add_column("Command")
            help_table.add_column("Description")

            for keyword in self._available_keywords:
                if keyword in self.keywords:
                    command_description = self.description
                else:
                    command_description = self._core_commands.get_command(keyword).description  # type: ignore # noqa: E501

                help_table.add_row(keyword, command_description)

            snutils.console.print(help_table)



import shlex
from typing import List, Optional, Tuple, Union

from prompt_toolkit import PromptSession
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.history import InMemoryHistory

from snadra._core.commands import Commands
import snadra._utils as snutils


class CommandParser:
    """
    Responsible for handling the commands entered.

    Maps each command and it's arguments to the desired action.
    """

    def __init__(self) -> None:
        self.commands = Commands()

    def setup_prompt(self) -> None:  # pragma: no cover
        """
        See Notes section.

        Notes
        -----
        The only reason for this being in a seperate function is that it changes
        the `sys.stdout` and `sys.stderr` which disturbes `pytest`.
        """
        self.prompt: "PromptSession[str]" = PromptSession(
            "snadra > ",
            auto_suggest=AutoSuggestFromHistory(),
            history=InMemoryHistory(),
        )

    def run(self) -> None:  # pragma: no cover # TODO: Remove this pragma
        """
        The main loop.

        This is an infitine loop, until the user decides to exit.
        """
        self.running = True
        while self.running:
            try:
                line = self.prompt.prompt().strip()
                if line == "":
                    continue
                self.dispatch_line(line)
            except EOFError:
                self.running = False
            except KeyboardInterrupt:
                continue
            except Exception:
                # Unexpected errors, we catch them so the application won't crash.
                snutils.console.print_exception(width=None, show_locals=True)
                continue

    def dispatch_line(self, line: str) -> None:
        """
        Execute each command that was entered to the console.

        Parameters
        ----------
        line : str
            The full command (including arguments) as a string.
        """
        try:
            argv, pline = CommandParser._parse_line(line=line)  # type: ignore
        except TypeError:
            return

        if self.commands.is_valid_keyword(argv[0]):
            # NOTE: Here is where we initialize the command
            command = self.commands.get_command(argv[0])()  # type: ignore
        else:
            snutils.console.log(f"[red]Error[/red]: {repr(argv[0])} unknown command")
            return

        args: Union[str, List[str]] = argv[1:]
        args = [arg.encode("utf-8").decode("unicode_escape") for arg in args]

        try:
            if command.parser:
                args = command.parser.parse_args(args)
            else:
                args = pline
            command.run(args)
        except SystemExit:
            snutils.console.log("Incorrect arguments")
            return

    @staticmethod
    def _parse_line(line: str) -> Optional[Tuple[List[str], str]]:
        """
        Parameters
        ----------
        line : str
            The full command (including arguments).

        Returns
        -------
        List[str]
            Line parsed as a list. (with shlex)
        str

        Raises
        ------
        ValueError
            If could not parse the line correctly.

        See Also
        --------
        shlex.split

        Examples
        --------
        >>> line = "command --target 127.0.0.1 --port 80"
        >>> CommandParser._parse_line(line)
        (\
['command', '--target', '127.0.0.1', '--port', '80'], \
'command --target 127.0.0.1 --port 80')
        """
        line = line.strip()
        if line == "":
            return None

        try:
            argv = shlex.split(line)
        except ValueError as err:
            snutils.console.log(f"[red]Error[/red]: {err.args[0]}")
            return None

        pline = f"{argv[0]} ".join(line.split(f"{argv[0]} "))
        return (argv, pline)



import importlib
from importlib.machinery import SOURCE_SUFFIXES
import importlib.util
import pathlib
import sys
from typing import (
    TYPE_CHECKING,
    Dict,
    FrozenSet,
    Iterable,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

import snadra._utils as snutils

if TYPE_CHECKING:
    import os
    import types

    from snadra._core.base import CommandDefinition
    from snadra._typing import StrPath


class Commands:
    """
    Holds all the relevant commands attributes.

    Parameters
    ----------
    path : snadra._typing.StrPath, optional.
        Path to the directory with the commands to load.
        If not specified, the snadra's core commands directory is being loaded.
        Sequence containing strings of paths to the command directories.
    skip : Union[Sequence[str], Set[str], FrozenSet[str]]], optional.
        Module names to skip.

    Notes
    -----
    The module names inside `skip`, should not be with a file suffix.
    """

    __slots__ = {"_commands_dict", "_path", "_skip"}

    def __init__(
        self,
        path: Optional["StrPath"] = None,
        *,
        skip: Optional[Union[Sequence[str], Set[str], FrozenSet[str]]] = None,
    ) -> None:
        if path is None:
            self._path = pathlib.Path(__file__).parent.resolve()
        else:
            self._path = pathlib.Path(path)

        if skip:
            self._skip = skip
        else:
            self._skip = frozenset({"__init__"})

        fetched_modules = Commands._fetch_modules(
            file_paths=Commands.iter_dir(path=self._path, skip=self._skip)
        )
        self._commands_dict: Dict[str, "CommandDefinition"] = {
            keyword: module.Command  # type: ignore
            for keyword, module in Commands._module_aliases(
                fetched_modules=fetched_modules
            )
        }

    @staticmethod
    def _fetch_modules(
        file_paths: Iterable["os.PathLike"],
    ) -> Iterable["types.ModuleType"]:
        """
        Get all modules from an iterable of file paths.

        Parameters
        ----------
        file_paths : Iterable[:class:`os.PathLike`]
            Iterable of file paths of python modules, that will be loaded.

        Yields
        ------
        :class:`types.ModuleType`
            Module that contains a `Command` class.

        Notes
        -----
        Skipping already loaded modules.
        """
        for path in file_paths:
            module_name = path.stem  # type: ignore
            if module_name in sys.modules:
                # TODO: Do we ever reach here?
                snutils.console.log(f"Skiping already loaded module {module_name}")
                continue

            module_spec = importlib.util.spec_from_file_location(module_name, path)
            module = importlib.util.module_from_spec(module_spec)
            module_spec.loader.exec_module(module)  # type: ignore
            yield module

    @staticmethod
    def _module_aliases(
        fetched_modules: Iterable["types.ModuleType"],
    ) -> Iterable[Tuple[str, "types.ModuleType"]]:
        """
        Exctracting the alias of each Command module.

        Parameters
        ----------
        fetched_modules : Iterable[:class:`types.ModuleType`]
            Generator of modules containing a `Command` class.

        Yields
        ------
        str
            Command alias.
        :class:`types.ModuleType`
            Module containing a `Command` class.
        """
        for module in fetched_modules:
            command = module.Command  # type: ignore
            for keyword in command.keywords:
                yield keyword, module

    @staticmethod
    def iter_dir(
        path: "os.PathLike",
        *,
        skip: Optional[Union[Sequence[str], Set[str], FrozenSet[str]]],
    ) -> Iterable["os.PathLike"]:
        """
        Iterating over a directory, skipping specified file names.

        Parameters
        ----------
        path : :class:`os.PathLike`
            Path of the direcory to iterate over.
        skip : Union[Sequence[str], Set[str], FrozenSet[str]], optional
            File names to skip.

        Yields
        ------
        :class:`os.PathLike`
            File paths that have not got skipped over.
        """
        # TODO(maybe): Add recursive for dirs?
        for child in path.iterdir():  # type: ignore
            if child.is_dir():
                continue
            if child.stem in skip:  # type: ignore
                continue
            if child.suffix not in SOURCE_SUFFIXES:
                continue
            yield child

    @property
    def keywords(self) -> Set[str]:
        """
        Get all the available keywords.

        Returns
        -------
        Set[str]
            All the available keywords.
        """
        return set(self._commands_dict.keys())

    def get_command(self, keyword: str) -> Optional["CommandDefinition"]:
        """
        Get the command that mapped to a keyword.

        Parameters
        ----------
        keyword : str
            Keyword to check.

        Returns
        -------
        Optional[:class:`CommandDefinition`]
            The command that is mapped to ``keyword``,
            if ``keyword`` is not mapped to any command, `None` is returned.
        """
        return self._commands_dict.get(keyword)

    def is_valid_keyword(self, keyword: str) -> bool:
        """
        Check if a given keyword is mapped to a valid command.

        Parameters
        ----------
        keyword : str
            Keyword to check.

        Returns
        -------
        bool
            Whether or not the keyword is mapped to a valid command.
        """
        return keyword in self._commands_dict



"""
Main application entry point.

This gets ran when you execute:

$ snadra

or

$ python -m snadra
"""
from snadra._core.parsers import CommandParser

if __name__ == "__main__":
    snadra_console = CommandParser()
    snadra_console.setup_prompt()
    snadra_console.run()
